@@ -6,9 +6,7 @@
    the Free Software Foundation in version 2 of the License.
 */
 
-#ifdef HAVE_CONFIG_H
-#  include <config.h>
-#endif
+#include <config.h>
 
 #include <stdio.h>
 #include <string.h>
@@ -50,5 +48,7 @@ int main(void)
 	return 1;
     }
 
+  free(myargv);
+
   return 0;
 }
